/*
 * Copyright © 2012 Linaro Limited
 *
 * This file is part of glcompbench.
 *
 * glcompbench is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * glcompbench is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with glcompbench.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authors:
 *  Alexandros Frantzis <alexandros.frantzis@linaro.org>
 *  Jesse Barker <jesse.barker@linaro.org>
 */
 
#include "log.h"
#include "render-object.h"
#include "shader-source.h"
#include "gl-headers.h"
#include "texture.h"

using LibMatrix::vec2;
using std::string;

bool
RenderObject::load_shaders_from_strings(Program &program,
    const string &vtx_shader, const string &frg_shader)
{
    program.init();

    Log::debug("Loading vertex shader:\n%s", vtx_shader.c_str());

    program.addShader(GL_VERTEX_SHADER, vtx_shader);
    if (!program.valid()) {
        Log::error("Failed to add vertex shader:\n  %s\n", "RenderObject: ",
            program.errorMessage().c_str());
        program.release();
        return false;
    }

    Log::debug("Loading fragment shader:\n%s", frg_shader.c_str());

    program.addShader(GL_FRAGMENT_SHADER, frg_shader);
    if (!program.valid()) {
        Log::error("Failed to add fragment shader:\n  %s\n",
            program.errorMessage().c_str());
        program.release();
        return false;
    }

    program.build();
    if (!program.ready()) {
        Log::error("Failed to link program:  %s\n",
            program.errorMessage().c_str());
        program.release();
        return false;
    }

    return true;
}

void
RenderObject::init(Program& default_program)
{
    default_program_ = &default_program;

    // Create a texture to use as a render target
    glGenTextures(1, &texture_);
    glBindTexture(GL_TEXTURE_2D, texture_);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, size_.x(), size_.y(), 0,
                 GL_RGBA, GL_UNSIGNED_BYTE, 0);

    // Create a FBO
    glGenFramebuffers(1, &fbo_);
    glBindFramebuffer(GL_FRAMEBUFFER, fbo_);
    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0,
                           GL_TEXTURE_2D, texture_, 0);

    glBindFramebuffer(GL_FRAMEBUFFER, 0);

    texture_contents_invalid_ = true;
}

void
RenderObject::release()
{
    // Release resources
    if (texture_ != 0)
    {
        glDeleteTextures(1, &texture_);
        texture_ = 0;
    }
    if (fbo_ != 0)
    {
        glDeleteFramebuffers(1, &fbo_);
        fbo_ = 0;
    }
}

void
RenderObject::make_current()
{
    glBindFramebuffer(GL_FRAMEBUFFER, fbo_);
    glViewport(0, 0, size_.x(), size_.y());
}

void
RenderObject::resize(const vec2& size)
{
    /// Recreate the backing texture with correct size
    if (size_ != size) {
        size_ = size;
        glBindTexture(GL_TEXTURE_2D, texture_);
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, size_.x(), size_.y(), 0,
                     GL_RGBA, GL_UNSIGNED_BYTE, 0);
        texture_contents_invalid_ = true;
    }

    if (texture_contents_invalid_) {
        clear();
        texture_contents_invalid_ = false;
    }
}

void
RenderObject::clear()
{
    make_current();
    glClear(GL_COLOR_BUFFER_BIT);
}

void
RenderObject::render_to(RenderObject& target)
{
    render_to(target, *default_program_);
}

void
RenderObject::render_to(RenderObject& target, Program& program)
{
    vec2 anchor(pos_);
    vec2 ll(pos_ - anchor);
    vec2 ur(pos_ + size_ - anchor);

    /* Calculate new position according to rotation value */
    GLfloat position[2 * 4] = {
        rotate_x(ll.x(), ll.y()) + anchor.x(), rotate_y(ll.x(), ll.y()) + anchor.y(),
        rotate_x(ur.x(), ll.y()) + anchor.x(), rotate_y(ur.x(), ll.y()) + anchor.y(),
        rotate_x(ll.x(), ur.y()) + anchor.x(), rotate_y(ll.x(), ur.y()) + anchor.y(),
        rotate_x(ur.x(), ur.y()) + anchor.x(), rotate_y(ur.x(), ur.y()) + anchor.y(),
    };

    /* Normalize position and write back to array */
    for (unsigned int i = 0; i < 4; i++) {
        unsigned int idx1(2 * i);
        unsigned int idx2(2 * i + 1);
        const vec2& v2(target.normalize_position(vec2(position[idx1], position[idx2])));
        position[idx1] = v2.x();
        position[idx2] = v2.y();
    }

    static const GLfloat texcoord[2 * 4] = {
        0.0, 0.0,
        1.0, 0.0,
        0.0, 1.0,
        1.0, 1.0,
    };

    target.make_current();

    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, texture_);
    draw_quad_with_program(position, texcoord, program);
}

void
RenderObject::draw_quad_with_program(const float *position,
    const float *texcoord, Program &program)
{
    int pos_index = program["position"].location();
    int tex_index = program["texcoord"].location();

    program.start();

    glEnableVertexAttribArray(pos_index);
    glEnableVertexAttribArray(tex_index);
    glVertexAttribPointer(pos_index, 2,
                          GL_FLOAT, GL_FALSE, 0, position);
    glVertexAttribPointer(tex_index, 2,
                          GL_FLOAT, GL_FALSE, 0, texcoord);

    glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);

    glDisableVertexAttribArray(tex_index);
    glDisableVertexAttribArray(pos_index);

    program.stop();
}

RenderClearImage::RenderClearImage(const string& image_file) :
    RenderObject(), background_texture_name_(image_file),
    background_texture_(0), background_is_image_(false),
    texcoord_(texcoord_window_)
{
    if (!background_texture_name_.empty())
    {
        background_is_image_ = true;
        texcoord_ = texcoord_image_;
    }
}

RenderClearImage::RenderClearImage() :
    RenderObject(),
    background_texture_(0), background_is_image_(false),
    texcoord_(texcoord_window_)
{
}

void
RenderClearImage::init(Program& program)
{
    RenderObject::init(program);
    if (background_is_image_)
    {
        // Load the image into a texture.
        Texture::load(background_texture_name_,
                      &background_texture_, GL_LINEAR, GL_LINEAR, 0);
    }
}

void
RenderClearImage::set_background(unsigned int background_texture)
{
    background_texture_ = background_texture;
}

void
RenderClearImage::clear()
{
    make_current();
    glClear(GL_COLOR_BUFFER_BIT);

    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, background_texture_);

    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    draw_quad_with_program(position_, texcoord_, *default_program_);
    glDisable(GL_BLEND);
}

const float RenderClearImage::position_[8] = {
    -1.0, -1.0,
     1.0, -1.0,
    -1.0,  1.0,
     1.0,  1.0,
};
const float RenderClearImage::texcoord_image_[8] = {
    0.0, 0.0,
    1.0, 0.0,
    0.0, 1.0,
    1.0, 1.0,
};
const float RenderClearImage::texcoord_window_[8] = {
    0.0, 1.0,
    1.0, 1.0,
    0.0, 0.0,
    1.0, 0.0,
};

